from flask import render_template, send_from_directory

from . import main


@main.route("/")
def index():
    """Homepage"""
    return render_template("index.html")


@main.route("/401")
def unauthorized():
    """Unauthorized Access"""
    return render_template("401.html")


@main.route("/views/id/<id>")
@main.route("/views//id/<id>")
def get_sample(id):
    return "<h3> this is id: " + id + "</h3>"


@main.route("/views/<path:path>")
@main.route("/views//<path:path>")
def send_view(path):
    return render_template("views/" + path)


@main.route("/static/<path:path>")
def send_js(path):
    return send_from_directory("static", path)



"""
API endpoint for groups
"""

from flask_jwt_extended import jwt_required

from seamm_dashboard import authorize
from seamm_dashboard.models import Group, GroupSchema


@jwt_required(optional=True)
@authorize.has_role("admin", "group manager")
def get_groups():

    groups = Group.query.all()

    # usernames = [ [ user.username for user in group.users ] for group in groups ]

    group_schema = GroupSchema(many=True)

    groups = group_schema.dump(groups)

    # for i, group in enumerate(groups):
    #    group["usernames"] = usernames[i]

    return groups, 200



from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField


class SoftwareForm(FlaskForm):
    name = StringField("What is your name?")
    domain = StringField("Which domain")
    submit = SubmitField("Submit")



"""
import_jobs.py

Functions to import jobs into the database
"""
import logging
import os
import sqlite3

from . import Flowchart, Job, Project
from .util import process_flowchart, process_job, file_owner

from seamm_dashboard import db

logger = logging.getLogger(__name__)


def import_jobs(location):
    """Import all the projects and jobs at <location>.

    <location> should be the path to the 'projects' directory in a datastore.
    All subdirectories will be added as projects, and the jobs within them
    added also.

    Parameters
    ----------
    location : str or path
        The projects directory in a datastore.

    Returns
    -------
    (n_projects, n_jobs) : int, integer
        The number of projects and jobs added to the database.
    """
    n_projects = 0
    n_added_projects = 0
    n_jobs = 0
    n_added_jobs = 0
    for potential_project in os.listdir(location):
        potential_project = os.path.join(location, potential_project)

        if os.path.isdir(potential_project):
            n_projects += 1
            project_name = os.path.basename(potential_project)
            logger.debug("Adding project {}".format(project_name))
            project, added = add_project(potential_project, project_name)
            if added:
                n_added_projects += 1

            for potential_job in os.listdir(potential_project):

                potential_job = os.path.join(potential_project, potential_job)

                if os.path.isdir(potential_job):
                    job_name = os.path.basename(potential_job)
                    logger.debug("       job {}".format(job_name))
                    try:
                        job, added = add_job(potential_job, job_name, project)
                    except sqlite3.IntegrityError as e:
                        logger.warning("Adding job {} failed: {}".format(job_name, e))
                    if job is None:
                        logger.debug("         was not a job!")
                    else:
                        n_jobs += 1
                        if added:
                            n_added_jobs += 1

    return (n_projects, n_added_projects, n_jobs, n_added_jobs)


def add_flowchart(flowchart_path, project):
    """Parse flowchart and add to data store if nedded.

    Analyze a flowchart and add to the database if it doesn't already exist. In
    either case return the Flowchart object.

    Parameters
    ----------
    flowchart_path : str
        The path to the flowchart to be added to the data store.
    project : Project
        The project object the flowchart belongs to.

    Returns
    -------
    Flowchart object.
    """

    # Analyze the flowchart given the path
    flowchart_info = process_flowchart(flowchart_path)

    flowchart = (
        db.session.query(Flowchart).filter_by(id=flowchart_info["id"]).one_or_none()
    )

    if flowchart is None:
        user, group = file_owner(flowchart_path)
        flowchart = Flowchart(owner_id=user, group_id=group, **flowchart_info)
        flowchart.projects.append(project)
        db.session.add(flowchart)
        db.session.commit()
    elif project not in flowchart.projects:
        flowchart.projects.append(project)
        db.session.commit()

    return flowchart


def add_job(job_path, job_name, project):
    """Add a job to the datastore. A unique job is based on directory location.

    If job_path does not have a .flow file, it is skipped and not added to the
    datastore.

    Parameters
    ----------
    job_path : str
        The directory containing the job to be added to the datastore.
    job_name : str
        The text name of the job, usually 'Job_xxxxxx'
    project : models.Project
        The Project object in the database
    """

    job_info = process_job(job_path)

    if job_info:
        flowchart_path = job_info.pop("flowchart_path")

        # Check if job is in DB
        found = db.session.query(Job).filter_by(path=job_info["path"]).one_or_none()

        if found is None:
            user, group = file_owner(job_path)
            job = Job(owner_id=user, group_id=group, **job_info)
            job.projects.append(project)

            add_flowchart(flowchart_path, project)

            db.session.add(job)
            db.session.commit()
            return job, True
        else:
            return found, False
    else:
        print(
            "No job found in directory {}. No job added to data store".format(job_path)
        )
        return None, False


def add_project(project_path, project_name):
    """
    Add a project to datastore.
    """

    # Check if in DB
    found = (
        db.session.query(Project)
        .filter_by(name=project_name, path=project_path)
        .one_or_none()
    )

    if found is None:
        user, group = file_owner(project_path)
        project = Project(
            path=project_path, name=project_name, owner_id=user, group_id=group
        )
        db.session.add(project)
        db.session.commit()
        return project, True
    else:
        return found, False



"""
Tests for the API (admin user)
"""


import os
import json

import pytest


@pytest.mark.skip
@pytest.mark.xfail
def test_delete_job(admin_client, project_directory):
    """Check delete method of api/jobs/{jobID}

    This is currently XFAIL. With jobs being a primary key in association tables,
    this is more complicated to delete
    """

    csrf_token = admin_client[1]
    admin_client = admin_client[0]

    csrf_headers = {"X-CSRF-TOKEN": csrf_token}

    expected_path = os.path.join(project_directory, "Job_000002")

    assert os.path.exists(expected_path)

    response = admin_client.delete("api/jobs/2", headers=csrf_headers)

    assert response.status_code == 200

    assert not os.path.exists(expected_path)


def test_get_users(admin_client):
    """Check get method of api/users on admin client"""

    admin_client = admin_client[0]

    response = admin_client.get("api/users")

    assert response.status_code == 200

    assert len(response.json) == 3


def test_add_user(admin_client):
    """Check post method of api/users on admin client."""
    csrf_token = admin_client[1]
    admin_client = admin_client[0]

    csrf_headers = {"X-CSRF-TOKEN": csrf_token}

    new_user = {
        "username": "waffles",
        "password": "waffles!",
        "roles": ["group manager"],
    }

    response = admin_client.post(
        "api/users",
        data=json.dumps(new_user),
        content_type="application/json",
        headers=csrf_headers,
    )

    assert response.status_code == 201, response.data



"""
API endpoint for groups
"""

from flask_jwt_extended import jwt_required

from seamm_dashboard import authorize
from seamm_dashboard.models import Role, RoleSchema


@jwt_required(optional=True)
@authorize.has_role("admin", "group manager")
def get_roles():

    roles = Role.query.all()

    role_schema = RoleSchema(many=True)

    roles = role_schema.dump(roles)

    return roles, 200



"""
API calls for projects
"""

from flask import Response
from flask_jwt_extended import jwt_required
from sqlalchemy import and_

from seamm_dashboard.models import Project, ProjectSchema, Job, JobSchema
from seamm_dashboard import authorize

__all__ = ["get_projects", "get_project", "get_project_jobs"]


@jwt_required(optional=True)
def get_projects(description=None, limit=None):

    # If limit is not set, set limit to all jobs in DB.
    if limit is None:
        limit = Project.query.count()
    if description is not None:
        projects = Project.query.filter(
            and_(Project.authorized("read"), Project.description.contains(description))
        ).limit(limit)
    else:
        projects = Project.query.filter(Project.authorized("read")).limit(limit)

    projects_schema = ProjectSchema(many=True)

    return projects_schema.dump(projects), 200


@jwt_required(optional=True)
def get_project(id):
    """
    Function for api endpoint api/projects/{id}

    Parameters
    ----------
    id : the ID of the project to return
    """
    project = Project.query.get(id)

    if project is None:
        return Response(status=404)

    if not authorize.read(project):
        return Response("You are not authorized to access this content.", status=401)

    project_schema = ProjectSchema(many=False)
    return project_schema.dump(project), 200


@jwt_required(optional=True)
def get_project_jobs(id):
    """
    Function for api endpoint api/projects/{id}/jobs. Get jobs associated with a
    project.

    Parameters
    ----------
    id : the ID of the project to return
    """

    project = Project.query.get(id)

    if project is None:
        return Response(status=404)

    if not authorize.read(project):
        return Response("You are not authorized to access this content.", status=401)

    jobs = []
    for job in project.jobs:
        jobs.append(Job.query.get(job.id))

    jobs_schema = JobSchema(many=True)

    return jobs_schema.dump(jobs), 200



from flask_wtf import FlaskForm
from wtforms import (
    StringField,
    PasswordField,
    SubmitField,
    SelectMultipleField,
    BooleanField,
)
from wtforms.fields.html5 import EmailField
from wtforms.validators import DataRequired, Length, Email, Regexp, EqualTo
from wtforms import ValidationError
from seamm_dashboard.models import User, Group


def _validate_group(self, field):
    if Group.query.filter(Group.name == field.data).first():
        raise ValidationError(
            f"Group name '{field.data}' already in use. Please pick a different group "
            "name."
        )


def _validate_user_delete(self, field):
    raise ValidationError("Input username does not match user ID.")


def _validate_group_delete(self, field):
    raise ValidationError("Input group name does not match group ID.")


def _validate_username(self, field):
    if User.query.filter(User.username == field.data).first():
        raise ValidationError(
            f"Username {field.data} already in use. Please pick a different username"
        )


def _validate_email(self, field):
    if User.query.filter(User.email == field.data).first():
        raise ValidationError(
            f"Email address {field.data} already in use. Please pick a different email "
            "address."
        )


def _password_none_or_usual(self, field):
    """
    This validator is for the manage user form. Either the password is not changed
    (len 0), or the password is changed and should meet the usual length requirement.
    """
    if 0 < len(field.data) < 7:
        raise ValidationError("Passwords must be at least 7 characters in length.")


# Common username field
_username = StringField(
    "Username",
    validators=[
        _validate_username,
        DataRequired(),
        Length(3, 64),
        Regexp(
            "^[A-Za-z][A-Za-z0-9_.]*$",
            0,
            "Usernames must have only letters, numbers, dots or " "underscores",
        ),
    ],
)


class CreateUsernamePasswordForm(FlaskForm):
    """
    A subform for creating a new username and password.
    """

    username = _username

    password2 = PasswordField("Confirm password", validators=[DataRequired()])

    password = PasswordField(
        "Password",
        validators=[
            DataRequired(),
            Length(min=7),
            EqualTo("password2", message="Passwords must match."),
        ],
    )


class EditUsernamePasswordForm(FlaskForm):
    """
    A subform for editing username and password.
    """

    username = _username

    password = PasswordField(
        "Password",
        validators=[
            _password_none_or_usual,
            EqualTo("password2", message="Passwords must match."),
        ],
    )

    password2 = PasswordField("Confirm Password")


class ContactInformationForm(FlaskForm):
    """
    A form for adding or updating contact information.
    """

    first_name = StringField("First Name", validators=[Length(2, 64)])

    last_name = StringField("Last Name", validators=[Length(2, 64)])

    email = EmailField(
        "Email Address",
        validators=[
            DataRequired(),
            Email(),
            _validate_email,
        ],
    )


class CreateUserForm(CreateUsernamePasswordForm, ContactInformationForm):
    """
    Form for adding or updating a user
    """

    roles = SelectMultipleField("User Roles", choices=[])

    groups = SelectMultipleField("User Groups", choices=[])

    submit = SubmitField("Create New User")


class ManageUserFormAdmin(EditUsernamePasswordForm, ContactInformationForm):
    """
    Form for adding or updating a user
    """

    roles = SelectMultipleField("User Roles", choices=[])

    groups = SelectMultipleField("User Groups", choices=[])

    submit = SubmitField("Update User Information")


class EditGroupForm(FlaskForm):
    """
    Form for adding or editing a group
    """

    group_name = StringField(
        "Group Name", validators=[Length(2, 64), DataRequired(), _validate_group]
    )

    group_members = SelectMultipleField("Group Members", choices=[])

    submit = SubmitField("Submit")


class DeleteUserForm(FlaskForm):
    """
    Form for deleting a user.
    """

    username = _username
    confirm = BooleanField("Confirm")
    submit = SubmitField("Delete User")


class DeleteGroupForm(FlaskForm):
    """
    Form for deleting a user.
    """

    group_name = StringField("Group Name", validators=[Length(2, 64), DataRequired()])
    confirm = BooleanField("Confirm")
    submit = SubmitField("Delete Group")



"""
API calls for the status
"""
import logging

from seamm_dashboard.models import Project, Job, Flowchart, RoleSchema

from sqlalchemy import and_

from flask_jwt_extended import get_current_user, jwt_required

logger = logging.getLogger("__file__")

__all__ = ["status"]


@jwt_required(optional=True)
def status():
    """The status of the dashboard.

    Currently always 'running', but in the future other
    values may be added.

    """
    # users = User.query.all()
    # user_schema = UserSchema(many=True)
    # users = user_schema.dump(users)

    current_user = get_current_user()

    try:
        username = current_user.username
    except AttributeError:
        username = "Anonymous User"

    if username != "Anonymous User":
        user_id = current_user.id
        roles = current_user.roles
        role_schema = RoleSchema(many=True)
        roles = role_schema.dump(roles)

        user_roles = []
        for role in roles:
            user_roles.append(role["name"])
    else:
        user_roles = []
        user_id = None

    # Get information about jobs, projects, flowcharts
    num_jobs = Job.query.filter(Job.authorized("read")).count()
    num_jobs_running = Job.query.filter(
        and_(Job.status == "running", Job.authorized("read"))
    ).count()
    num_jobs_finished = Job.query.filter(
        and_(Job.status == "finished", Job.authorized("read"))
    ).count()
    num_jobs_queued = Job.query.filter(
        and_(Job.status == "running", Job.authorized("submitted"))
    ).count()
    num_flowcharts = Flowchart.query.filter(Flowchart.authorized("read")).count()
    num_projects = Project.query.filter(Project.authorized("read")).count()

    # Build return json
    status = {
        "status": "running",
        "user id": user_id,
        "username": username,
        "roles": user_roles,
        "jobs": {
            "total": num_jobs,
            "running": num_jobs_running,
            "finished": num_jobs_finished,
            "queued": num_jobs_queued,
        },
        "flowcharts": num_flowcharts,
        "projects": num_projects,
    }

    return status, 200



"""
API calls for jobs.
"""

import os
import shutil
from datetime import datetime
import fasteners
import hashlib
import json
import logging
from pathlib import Path
import re
import urllib.parse

from marshmallow import ValidationError
from sqlalchemy import and_
from flask import send_from_directory, Response
from flask_jwt_extended import jwt_required

from seamm_dashboard import db, datastore, authorize, options
from seamm_dashboard.models import User, Project, Job, JobSchema, Flowchart
from seamm_dashboard.models import FlowchartSchema

logger = logging.getLogger("__file__")

__all__ = [
    "get_jobs",
    "get_job",
    "get_job_files",
    "add_job",
    "update_job",
    "delete_job",
]

file_icons = {
    "graph": "fas fa-chart-line",
    "csv": "fas fa-table",
    "flow": "fas fa-project-diagram",
    "other": "far fa-file-alt",
    "folder": "far fa-folder-open",
    "pdb": "fas fa-cubes",
    "mmcif": "fas fa-cubes",
    "cif": "fas fa-cubes",
}


@jwt_required(optional=True)
def get_jobs(createdSince=None, createdBefore=None, limit=None):
    """
    Function for API endpoint /api/jobs

    Parameters
    ----------
    createdSince: str
        Return jobs created after this date. Must be in format M-D-YYYY where M all
        numbers are integers.
    createdBefore:str
        Return jobs created before this date. Must be in format M-D-YYYY where M all
        numbers are integers.
    limit: int
        The maximum number of jobs to return.

    """

    # Handle dates
    if createdSince is not None:
        createdSince = datetime.strptime(createdSince, "%m-%d-%Y")
    else:
        # Basically all times
        createdSince = datetime.strptime("1-1-0001", "%m-%d-%Y")

    if createdBefore is not None:
        createdBefore = datetime.strptime(createdBefore, "%m-%d-%Y")
    else:
        # Up to now.
        createdBefore = datetime.utcnow()

    # If limit is not set, set limit to all jobs in DB.
    if limit is None:
        limit = Job.query.count()

    jobs = Job.query.filter(
        and_(Job.submitted > createdSince, Job.submitted < createdBefore),
        Job.authorized("read"),
    ).limit(limit)

    jobs_schema = JobSchema(many=True)

    return jobs_schema.dump(jobs), 200


def get_job_id(filename):
    """Get the next job id from the given file.

    This uses the fasteners module to provide locking so that
    only one job at a time can access the file, so that the job
    ids are unique and monotonically increasing.
    """

    filename = os.path.expanduser(filename)

    lock_file = filename + ".lock"
    lock = fasteners.InterProcessLock(lock_file)
    locked = lock.acquire(blocking=True, timeout=5)

    if locked:
        if not os.path.isfile(filename):
            job_id = 1
            with open(filename, "w") as fd:
                fd.write("!MolSSI job_id 1.0\n")
                fd.write("1\n")
            lock.release()
        else:
            with open(filename, "r+") as fd:
                line = fd.readline()
                pos = fd.tell()
                if line == "":
                    lock.release()
                    raise EOFError("job_id file '{}' is empty".format(filename))
                line = line.strip()
                match = re.fullmatch(r"!MolSSI job_id ([0-9]+(?:\.[0-9]+)*)", line)
                if match is None:
                    lock.release()
                    raise RuntimeError(
                        "The job_id file has an incorrect header: {}".format(line)
                    )
                line = fd.readline()
                if line == "":
                    lock.release()
                    raise EOFError("job_id file '{}' is truncated".format(filename))
                try:
                    job_id = int(line)
                except TypeError:
                    raise TypeError(
                        "The job_id in file '{}' is not an integer: {}".format(
                            filename, line
                        )
                    )
                job_id += 1
                fd.seek(pos)
                fd.write("{:d}\n".format(job_id))
    else:
        raise RuntimeError("Could not lock the job_id file '{}'".format(filename))

    return job_id


@jwt_required(optional=True)
def add_job(body):
    """Add a new job to the queue.

    Parameters
    ----------
    body : dict
        The description of the job.

    Returns
    -------
    The job id (integer)
    """

    logger.debug("Adding a job. Items in the body are:")
    for key, value in body.items():
        logger.debug("  {:15s}: {}".format(key, str(value)[:20]))

    # Get the project
    project_name = body.pop("project")
    project = db.session.query(Project).filter_by(name=project_name).one_or_none()
    if project is None:
        return "There is no project '{}'".format(project_name), 403

    logger.debug("  project = {}".format(project.id))

    # Get the user...
    username = body.pop("username")
    user = db.session.query(User).filter_by(username=username).one_or_none()
    if user is None:
        return "User '{}' is not registered".format(username), 403
    user_id = user.id
    group_id = user.groups[0].id
    body["owner"] = user_id
    body["group"] = group_id

    logger.debug("  user {} is {}:{}".format(username, user_id, group_id))

    # Get the flowchart and put it in place
    flowchart_text = body.pop("flowchart")
    flowchart_data = {
        "id": hashlib.md5(flowchart_text.encode("utf-8")).hexdigest(),
        "owner_id": user_id,
        "group_id": group_id,
        "description": "No description given.",
    }

    # Validate and deserialize the flowchart data
    flowchart_schema = FlowchartSchema(many=False)
    try:
        flowchart_schema.load(flowchart_data, session=db.session)
    except ValidationError as err:
        logger.error("ValidationError (flowchart): {}".format(err.messages))
        logger.info("   valid data: {}".format(err.valid_data))
        return err.messages, 422

    # See if it already in the db, if not, add it
    flowchart_is_new = False
    flowchart = (
        db.session.query(Flowchart).filter_by(id=flowchart_data["id"]).one_or_none()
    )
    if flowchart:
        if project not in flowchart.projects:
            flowchart.projects.append(project)
            db.session.commit()

    else:
        flowchart = Flowchart(
            **flowchart_data,
            text=flowchart_text,
            json="\n".join(flowchart_text.splitlines()[2:]),
        )
        flowchart.projects.append(project)
        db.session.add(flowchart)
        db.session.commit()
        flowchart_is_new = True

    # Validate and deserialize the job data
    body["status"] = "Submitted"
    body["flowchart_id"] = flowchart.id

    job_schema = JobSchema(many=False)
    try:
        job_schema.load(body, session=db.session)
    except ValidationError as err:
        logger.error("ValidationError (job): {}".format(err.messages))
        logger.info("   valid data: {}".format(err.valid_data))
        return err.messages, 422

    # Need objects, not ids?
    # body["flowchart_id"] = flowchart
    body["owner"] = user
    body["group"] = user.groups[0]

    job = Job(**body)

    # Get the unique ID for the job...

    if options["job_id_file"] is None:
        job_id_file = os.path.join(datastore, "job.id")
    else:
        job_id_file = options["job_id_file"]
    job.id = get_job_id(job_id_file)

    # And the path
    project_path = Path(datastore).expanduser() / "projects" / project_name
    directory = project_path / "Job_{:06d}".format(job.id)
    print("Writing job files to " + str(directory))
    directory.mkdir(parents=True, exist_ok=True)
    job.path = str(directory)

    # Finally add the job to the database.
    db.session.add(job)
    job.projects.append(project)

    db.session.commit()

    path = directory / "flowchart.flow"
    with path.open("w") as fd:
        fd.write(flowchart_text)
    if flowchart_is_new:
        flowchart.path = str(path)

    db.session.commit()

    # Write the json data file for the job
    data = {
        "command line": "",
        "title": body["title"],
        "working directory": str(directory),
        "state": "Submitted",
        "projects": [project_name],
        "datastore": datastore,
        "job id": job.id,
    }
    path = directory / "job_data.json"
    with path.open("w") as fd:
        json.dump(data, fd, sort_keys=True, indent=3)

    return {"id": job.id}, 201, {"location": format("/jobs/{}".format(job.id))}


@jwt_required(optional=True)
def get_job(id):
    """
    Function for api endpoint api/jobs/{id}

    Parameters
    ----------
    id : the ID of the job to return
    """
    if not isinstance(id, int):
        return Response(status=400)

    job = Job.query.get(id)

    if job is None:
        return Response(status=404)

    if not authorize.read(job):
        return Response("You are not authorized to view this content.", status=401)

    job_schema = JobSchema(many=False)
    return job_schema.dump(job), 200


@jwt_required(optional=True)
def update_job(id, body):
    """
    Function to update jobs - endpoint api/jobs/{id}

    Parameters
    ----------
    id : int
        The ID of the job to update

    body : json
        The job information to update.
    """
    job = Job.query.get(id)

    if not job:
        return Response(status=404)

    if not authorize.update(job):
        return Response(status=401)

    for key, value in body.items():
        if key == "submitted" or key == "finished" or key == "started":
            if value:
                # This assumes the timestamp is in format from javascript Date.now()
                # https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date/now
                # Number of milliseconds since January 1, 1970
                value = datetime.fromtimestamp(value / 1000)
            else:
                value = None
        setattr(job, key, value)

    db.session.commit()

    return Response(status=201)


@jwt_required(optional=True)
@authorize.has_role("admin")
def delete_job(id):
    """
    Function for delete method of api endpoint api/jobs/{id}

    This api route removes the job from the DB and deletes the associated job files
    from disk

    Parameters
    ----------
    id : int
        The ID of the job to delete

    Returns
    -------
    status : int
        Response code for operation. 200 = successful, 404 = job not found.
    """
    job = Job.query.get(id)
    print("deleting job")

    if not job:
        return Response(status=404)
    else:
        path = job.path
        job_path = Path(path)

        # Remove job files if they exist
        if job_path.exists():
            shutil.rmtree(job_path)

        # Remove job info from DB
        db.session.delete(job)
        db.session.commit()

        return Response(status=200)


@jwt_required(optional=True)
def get_job_files(id, file_path=None):
    """
    Function for get method of api endpoint api/jobs/{id}/files. If the file_path
    parameter is used, this endpoint will send the file which is indicated by the path.

    Parameters
    ----------
    id : int
        the ID of the job to return

    file_path : string
        The encoded file path for the file to return.
    """

    job_info, status = get_job(id)

    if file_path is None:
        js_tree = []

        path = job_info["path"]

        base_dir = os.path.split(path)[1]

        js_tree.append(
            {
                "id": path,
                "parent": "#",
                "text": base_dir,
                "state": {
                    "opened": "true",
                    "selected": "true",
                },
                "icon": file_icons["folder"],
            }
        )

        for root, dirs, files in os.walk(path):
            parent = root

            for name in sorted(files):

                extension = name.split(".")[-1]

                encoded_path = urllib.parse.quote(os.path.join(root, name), safe="")

                js_tree.append(
                    {
                        "id": encoded_path,
                        "parent": parent,
                        "text": name,
                        "a_attr": {
                            "href": f"api/jobs/{id}/files?file_path={encoded_path}",
                            "class": "file",
                        },
                        "icon": [
                            file_icons[extension]
                            if extension in file_icons.keys()
                            else file_icons["other"]
                        ][0],
                    }
                )

            for name in sorted(dirs):
                js_tree.append(
                    {
                        "id": os.path.join(root, name),
                        "parent": parent,
                        "text": name,
                        "icon": file_icons["folder"],
                    }
                )
        return js_tree

    else:
        unencoded_path = urllib.parse.unquote(file_path)
        directory, file_name = os.path.split(unencoded_path)
        return send_from_directory(directory, path=file_name, as_attachment=True)



import logging
import os
from pathlib import Path
import time

import connexion

# from flask_admin import Admin
from flask_debugtoolbar import DebugToolbarExtension
from flask_bootstrap import Bootstrap
from flask_cors import CORS
from flask_mail import Mail
from flask_moment import Moment
from flask_sqlalchemy import SQLAlchemy
from flask_marshmallow import Marshmallow

from flask_authorize import Authorize

from .jwt_patch import flask_jwt_extended

from flask_jwt_extended import get_current_user

from .config import config
from .template_filters import replace_empty
from .setup_logging import setup_logging
from .setup_argparsing import options, parser


# Handle versioneer
from ._version import get_versions

__author__ = """Jessica Nash"""
__email__ = "janash@vt.edu"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions

# Ensure that the projects directory exists.
datastore = Path(options["datastore"]).expanduser().resolve()
datastore.mkdir(parents=True, exist_ok=True)

# Setup the logging, now that we know where the datastore is
setup_logging(str(datastore), options)
logger = logging.getLogger("dashboard")

# Two of the Flask options cannot be reset, and should (apparently) be
# handled with environment variables ... so if they are in the options
# set the correct environment variables. Carefully!

if "env" in options:
    if "FLASK_ENV" in os.environ and options["env"] != os.environ["FLASK_ENV"]:
        logger.warning(
            (
                "The environment variable FLASK_ENV is being overidden by "
                "the configuration option 'env' ({})"
            ).format(options["env"])
        )
    os.environ["FLASK_ENV"] = options["env"]
if "debug" in options:
    if "FLASK_DEBUG" in os.environ and options["debug"] != os.environ["FLASK_DEBUG"]:
        logger.warning(
            (
                "The environment variable FLASK_DEBUG is being overidden by "
                "the configuration option 'debug' ({})"
            ).format(options["debug"])
        )
    os.environ["FLASK_DEBUG"] = options["debug"]

# continue the setup
mail = Mail()
cors = CORS()

bootstrap = Bootstrap()

jwt = flask_jwt_extended.JWTManager()
authorize = Authorize(current_user=get_current_user)

moment = Moment()
toolbar = DebugToolbarExtension()

db = SQLAlchemy()
ma = Marshmallow()


def create_app(config_name=None):
    """Flask app factory pattern
    separately creating the extensions and later initializing"""

    conn_app = connexion.App(__name__, specification_dir="./")
    app = conn_app.app

    logger.info("")
    if config_name is not None:
        logger.info("Configuring from configuration " + config_name)
        app.config.from_object(config[config_name])

        options["initialize"] = False
        options["no_check"] = True
    else:
        # Report where options come from
        # parser = configargparse.get_argument_parser("dashboard")
        logger.info("Where options are set:")
        logger.info(60 * "-")
        for section, tmp in parser.get_options().items():
            origin = parser.get_origins(section)
            for key, value in tmp.items():
                logger.info(f"{key:<19} {origin[key]:<15} {value}")

        # Now set the options!
        logger.info("")
        logger.info("Configuration:")
        logger.info(60 * "-")
        for key, value in options.items():
            if key not in (
                "env",
                "debug",
                "initialize",
                "log_dir",
                "log_level",
                "console_log_level",
                "dashboard_configfile",
            ):
                key = key.upper()
                if isinstance(value, str):
                    value = value.replace("%datastore%", str(datastore))
                logger.info("\t{:>30s} = {}".format(key, value))
                app.config[key] = value

    logger.info("")

    logger.info(
        "Running in "
        + app.config["ENV"]
        + " mode with database "
        + app.config["SQLALCHEMY_DATABASE_URI"]
    )

    conn_app.add_api("swagger.yml")

    db.init_app(app)
    with app.app_context():
        if options["initialize"]:
            logger.info("Removing all previous jobs from the database.")
            db.drop_all()
        db.create_all()

        from .routes.auth import auth as auth_blueprint
        from .routes.main import main as main_blueprint
        from .routes.jobs import jobs as jobs_blueprint
        from .routes.flowcharts import flowcharts as flowchart_blueprint
        from .routes.projects import projects as project_blueprint
        from .routes.admin import admin as admin_blueprint

        from .routes.main import errors

        app.register_blueprint(auth_blueprint)
        app.register_blueprint(main_blueprint)
        app.register_blueprint(jobs_blueprint)
        app.register_blueprint(flowchart_blueprint)
        app.register_blueprint(project_blueprint)
        app.register_blueprint(admin_blueprint)

        app.register_error_handler(404, errors.not_found)

    # init
    mail.init_app(app)
    cors.init_app(app)
    bootstrap.init_app(app)
    authorize.init_app(app)
    jwt.init_app(app)
    moment.init_app(app)

    # jinja template
    app.jinja_env.filters["empty"] = replace_empty

    # Authorization configuration
    app.config["AUTHORIZE_DEFAULT_PERMISSIONS"] = dict(
        owner=["read", "update", "delete", "create", "manage"],
        group=["read", "update"],
        other=[""],
    )
    app.config["AUTHORIZE_ALLOW_ANONYMOUS_ACTIONS"] = True

    # Set application to store JWTs in cookies.
    app.config["JWT_TOKEN_LOCATION"] = ["cookies"]

    # Set the cookie paths
    # app.config["JWT_ACCESS_COOKIE_PATH"] = "/api"
    app.config["JWT_REFRESH_COOKIE_PATH"] = "/api/auth/token/refresh"

    # Cookie security
    app.config["JWT_COOKIE_SECURE"] = False
    app.config["JWT_COOKIE_CSRF_PROTECT"] = True
    app.config["JWT_CSRF_ACCESS_PATH"] = "/api/"

    # Add some default roles to the dashboard
    with app.app_context():
        from .models import Role

        role_names = ["user", "group manager", "admin"]

        for role_name in role_names:
            check = Role.query.filter_by(name=role_name).one_or_none()
            if not check:
                role = Role(name=role_name)
                db.session.add(role)
                db.session.commit()

        # Add an admin group and user if not present
        from .models import Group

        name = "admin"
        group = db.session.query(Group).filter_by(name=name).one_or_none()
        if group is None:
            group = Group(name=name)
            db.session.add(group)
            db.session.commit()

        from .models import User

        name = "admin"
        user = db.session.query(User).filter_by(username=name).one_or_none()
        if user is None:
            admin_role = db.session.query(Role).filter_by(name="admin").one_or_none()

            if admin_role is None:
                admin_role = Role(name="admin")

            user = User(username=name, password="admin", roles=[admin_role])
            user.groups.append(group)
            db.session.add(user)
            db.session.add(admin_role)
            db.session.add(group)
            db.session.commit()

        # And the current user also
        item = Path.home()
        # Get the group first
        name = item.group()
        group = db.session.query(Group).filter_by(name=name).one_or_none()
        if group is None:
            group = Group(name=name)
            db.session.add(group)
            db.session.commit()

        # and now the user
        name = item.owner()
        user = db.session.query(User).filter_by(username=name).one_or_none()
        if user is None:
            admin_role = db.session.query(Role).filter_by(name="admin").one_or_none()

            if admin_role is None:
                admin_role = Role(name="admin")

            user = User(username=name, password="default", roles=[admin_role])
            user.groups.append(group)
            db.session.add(user)
            db.session.add(admin_role)
            db.session.add(group)
            db.session.commit()

        # Add a default project if it does not exist.

        from .models import Project

        # Ensure that the directory exists
        projects = datastore / "projects"
        default = projects / "default"
        default.mkdir(parents=True, exist_ok=True)

        # Check if in DB
        if (
            db.session.query(Project)
            .filter_by(name="default", path=str(default))
            .one_or_none()
        ) is None:
            # Note this relies on the courrent user's id and group from above.
            project = Project(
                path=str(default), name="default", owner_id=user.id, group_id=group.id
            )
            db.session.add(project)
            db.session.commit()

    logger.info("")
    logger.info("Final configuration:")
    logger.info(60 * "-")
    for key, value in app.config.items():
        logger.info("\t{:>30s} = {}".format(key, value))
    logger.info("")

    if not options["no_check"]:
        # Ugly but avoids circular import.
        from .models.import_jobs import import_jobs

        t0 = time.perf_counter()
        with app.app_context():
            n_projects, n_added_projects, n_jobs, n_added_jobs = import_jobs(
                str(projects)
            )

            db.session.commit()

        t1 = time.perf_counter()
        logger.info(
            "Checked {} jobs and {} projects in {:.2f} s.".format(
                n_jobs, n_projects, t1 - t0
            )
        )
        if n_added_jobs > 0 or n_added_projects > 0:
            logger.info(
                "  added {} jobs and {} projects".format(n_added_jobs, n_added_projects)
            )

    logger.info(f"{app.url_map}")

    return app



from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField, SubmitField
from wtforms.fields.html5 import EmailField

from wtforms.validators import DataRequired, Length, EqualTo, Email

from seamm_dashboard.routes.admin.forms import _validate_email, _password_none_or_usual


class LoginForm(FlaskForm):
    username = StringField("Username", validators=[DataRequired(), Length(3, 64)])
    password = PasswordField("Password", validators=[DataRequired()])
    # remember_me = BooleanField("Keep me logged in")
    submit = SubmitField("Log In")


class ConfirmLogin(FlaskForm):
    """A login form which doesn't have a username field"""

    password = PasswordField("Password", validators=[DataRequired()])
    submit = SubmitField("Log In")


class UpdateAccountInfoForm(FlaskForm):
    """
    A form a user uses to update their own information. Users cannot change their
    username.
    """

    password2 = PasswordField("Confirm password", validators=[_password_none_or_usual])

    password = PasswordField(
        "Password",
        validators=[
            EqualTo("password2", message="Passwords must match."),
        ],
    )

    first_name = StringField("First Name", validators=[Length(2, 64)])

    last_name = StringField("Last Name", validators=[Length(2, 64)])

    email = EmailField(
        "Email Address",
        validators=[
            DataRequired(),
            Email(),
            _validate_email,
        ],
    )

    submit = SubmitField("Update User Information")



from .models import (  # noqa: F401
    Job,
    Flowchart,
    Project,
    User,
    Group,
    Role,
    UserJobAssociation,
    GroupJobAssociation,
    UserProjectAssociation,
    GroupProjectAssociation,
    GroupFlowchartAssociation,
)
from .models import (  # noqa: F401
    JobSchema,
    FlowchartSchema,
    ProjectSchema,
    UserSchema,
    GroupSchema,
    RoleSchema,
)
from .import_jobs import import_jobs  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""seamm-dashboard
The Web Dashboard for SEAMM (Simulation Environment for Atomistic and Molecular Simulations).
"""
import setuptools
import versioneer

with open("requirements_install.txt") as fd:
    requirements = fd.read()

if __name__ == "__main__":
    with open("requirements_install.txt") as fd:
        requirements = fd.read()

    setuptools.setup(
        name="seamm-dashboard",
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        description=__doc__.splitlines()[1],
        author="Jessica Nash",
        author_email="janash@vt.edu",
        url="https://github.com/molssi-seamm/seamm_dashboard.git",
        license="BSD-3C",
        packages=setuptools.find_packages(),
        # Required packages, pulls from pip if needed; do not use for Conda
        # deployment
        install_requires=requirements,
        include_package_data=True,
        extras_require={
            "tests": [],
        },
        tests_require=[],
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Natural Language :: English",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.8",
        ],
        zip_safe=False,
        entry_points={
            "console_scripts": [
                "seamm_dashboard=seamm_dashboard.results_dashboard:run",
                "seamm-dashboard=seamm_dashboard.results_dashboard:run",
            ],
        },
    )



from flask_wtf import FlaskForm
from wtforms import StringField, TextAreaField, SubmitField


class EditProject(FlaskForm):
    name = StringField("Project Title")
    notes = TextAreaField("Description")
    submit = SubmitField("Update Project")


class ManageProjectAccessForm(FlaskForm):

    submit = SubmitField("Update Project Access")
